/*
 * CG Madness - a Marble Madness clone
 * Copyright (C) 2007  Sven Reinck <sreinck@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "SandboxGame.hpp"

#include "ball/PlayersBall.hpp"

#include "process/Choose.hpp"
#include "process/MainProcess.hpp"
#include "process/ballcamera.hpp"

#include "screen/sandbox/ScreenSandboxLobby.hpp"
#include "screen/sandbox/ScreenSandboxEnd.hpp"
#include "screen/choose/ScreenChooseInfo.hpp"

#include "LevelLoader.hpp"

#include "SandboxProxy.hpp"
#include "server/ServerThread.hpp"
#include "net/Client.hpp"

#include "hw/keyboard.hpp"

#include "utils/Callback.hpp"

#include "environment/environment.hpp"

#include "text/text.hpp"

#include "macros.hpp"

#include GL_H

#include <cstdio>

SandboxGame::SandboxGame() :
	mMyID(GRAPPLE_EVERYONE), mMasterID(GRAPPLE_SERVER)
{
	gScreenSandboxLobby = new ScreenSandboxLobby(this);
	gScreenChooseInfo = new ScreenChooseInfo(METHOD_CALLBACK(SandboxGame, levelChosen));
	mChoose = new Choose(gScreenChooseInfo);
	gScreenEnd = new ScreenSandboxEnd(this);
}

SandboxGame::~SandboxGame()
{
	// empty
}

void SandboxGame::start(Process* previous, bool push)
{
	Game::start(previous, push);

	Main::pushState(gScreenSandboxLobby);
}

void SandboxGame::stop(Process* next)
{
	Game::stop(next);

	if (next != mChoose)
	{
		if (mServerThread)
		{
			mServerThread->stopServer();
		}

		mSandboxProxy = NULL;
		mClient = NULL;
	}
}

void SandboxGame::suspend()
{
	Game::suspend();

	mSandboxState = STATE_LOBBY;
	mStatusDirty = true;
}

void SandboxGame::resume()
{
	Game::resume();

	mSandboxState = STATE_RUNNING;
	changeMyState(PlayerInfo::STATE_PLAYING);
}

void SandboxGame::setPlayerName(const String& name)
{
	mPlayerName = name;
}

void SandboxGame::chooseLevel()
{
	Main::setState(mChoose, true);
}

void SandboxGame::startAllGames()
{
	int session = mStatus.nextSession();
	mSandboxProxy->resumeGame(GRAPPLE_EVERYONE, session);
}

void SandboxGame::endAllGames()
{
	mSandboxProxy->endGame(GRAPPLE_EVERYONE);
}

void SandboxGame::changeMyState(PlayerInfo::PlayerState newState)
{
	const PlayerMap& playerMap = mStatus.playerMap();

	PlayerMap::const_iterator iter = playerMap.find(mMyID);

	if (iter != playerMap.end())
	{
		mStatus.changeStatusState(mMyID, iter->second.time, newState);
		mSandboxProxy->changeState(GRAPPLE_EVERYONEELSE, iter->second.time, iter->second.state);
	}
}

void SandboxGame::gotoLobby()
{
	resetGame();
	mStatusDirty = true;

	if (mSandboxState == STATE_RUNNING)
	{
		Main::pushState(gScreenSandboxLobby);
	}
	else
	{
		Main::setState(gScreenSandboxLobby, false);
	}
	changeMyState(PlayerInfo::STATE_LOBBY);
}

void SandboxGame::leave()
{
	if (mSandboxState == STATE_RUNNING)
	{
		Main::pushState(gScreenEnd);
	}
	else
	{
		Main::setState(gScreenEnd, false);
	}
	changeMyState(PlayerInfo::STATE_ABORTED);
}

void SandboxGame::disconnect()
{
	Singleton<MainProcess> main;
	Main::setState(main, true);
}

void SandboxGame::levelChosen()
{
	Main::setState(this, true);

	mSandboxProxy->loadLevel(GRAPPLE_EVERYONEELSE, mLevelLoader->name());
}

void SandboxGame::startServer()
{
	mServerThread = new ServerThread();
	mServerThread->startServer();

	startClient("localhost");
}

void SandboxGame::startClient(const String& servername)
{
	mStatus.clear();

	mClient = new Client(PROGRAM_NAME, MACRO_STRING(PROTOCOL_VERSION), servername, mPlayerName);

	mSandboxProxy = new SandboxProxy(*mClient);

	mSandboxState = STATE_LOBBY;
	mConnectionState = STATE_CONNECTING;
	mReason = 0;
	mMaster = false;
	mStatusDirty = true;

	Main::setState(this, true);
}

void SandboxGame::update(float interval)
{
	switch (mSandboxState)
	{
	case STATE_LOBBY:
		mClient->processMessages(this);

		Game::update(interval);
		break;
	case STATE_RUNNING:
		if (mStatus.ticks() > mLastUpdate)
		{
			mStatusDirty = true;
		}

		mClient->processMessages(this);

		if (wasKeyPressed(KEY_ENTER))
		{
			sgoBall.explodeBall();
			mSandboxProxy->explode(GRAPPLE_EVERYONEELSE);
		}

		bool wasInactive = sgoBall.isInPieces() || !isBallActive();

		Game::update(interval);

		if (wasInactive && !(sgoBall.isInPieces() || !isBallActive()))
		{
			mStatus.incTrys(mMyID, mStatus.time());
			mSandboxProxy->newTry(GRAPPLE_EVERYONEELSE, mStatus.session(), mStatus.time());
		}

		if (sgoBall.hasHitGoal())
		{
			finishedGame();
		}

		break;
	}

	mStatus.update(interval);

	if (mStatusDirty)
	{
		if (!sgoBall.isInPieces())
		{
			mSandboxProxy->setPosition(GRAPPLE_EVERYONEELSE, mStatus.session(),
			    mStatus.time(), PlayersBall::sgoBall.pos(),
			    PlayersBall::sgoBall.velocity());
		}
		mLastUpdate = mStatus.ticks();
		mStatusDirty = false;
	}
}

void SandboxGame::draw() const
{
	Game::draw();

	FOREACH(mStatus.ballMap(), iter)
	{
		const NetworkBall& ball = iter->second.mBall;

#ifdef DEBUG_LASTTIME
		const SandboxStatus::BallMap& ballMap = mStatus.ballMap();

		SandboxStatus::BallMap::const_iterator iter2 = ballMap.find(iter->first);

		if (iter2 != ballMap.end())
		{
			glRasterPos3fv(&ball.pos().x);

			char str[20];

			sprintf(str, "%.3f", iter2->second.mLastTime);

			drawBitmapText(str);
		}
#endif

		ball.drawGameBall();
	}
}

void SandboxGame::drawHUD(float widthWindow, float heightWindow)
{
	if (mSandboxState == STATE_RUNNING)
	{
		int tenthSecond = (int) (mStatus.time() * 10.0f);
		float scale = 0.06f;
		float widthDefault = widthStrokeText("x:xx.x") * scale;

		char strTime[10];
		float width;
		float height;

		sprintf(strTime, "%d:%02d.%01d", tenthSecond / 600, tenthSecond / 10 % 60,
				tenthSecond % 10);

		width = widthStrokeText(strTime) * scale;
		height = scale;

		glColor3f(1.0f, 1.0f, 0.0f);

		glPushMatrix();
		{
			glTranslatef((widthWindow - widthDefault) / 2.0f, (heightWindow - height), 0.0f);
			glScalef(scale, scale, scale);

			drawStrokeThickText(strTime);
		}
		glPopMatrix();
	}
}

void SandboxGame::drawWaterReflection() const
{
	Game::drawWaterReflection();

	FOREACH(mStatus.ballMap(), iter)
	{
		const Ball& ball = iter->second.mBall;

		ball.drawGameBall();
	}
}

void SandboxGame::drawBallReflection() const
{
	Game::drawBallReflection();

	FOREACH(mStatus.ballMap(), iter)
	{
		const Ball& ball = iter->second.mBall;

		ball.drawGameBall();
	}
}

void SandboxGame::log(const String& log)
{
	printf("Client: %s\n", (const char*) log);
}

void SandboxGame::newUser(grapple_user id, bool me, const char* name)
{
	int number = (id & 0xffff) >> 1;
	int base = number % 6 + 1;
	int depth = number / 6;
	int scale = 0;

	int x = depth;

	while (x)
	{
		x >>= 1;
		scale++;
	}

	int one = 1 << scale;
	int frac = (one - 1) ^ ((depth << 1) ^ one);
	float f = (float) frac / one;

	float r = f * (base & 1);
	float g = f * ((base >> 1) & 1);
	float b = f * ((base >> 2) & 1);

	Color3 color(r, g, b);

	if (name)
	{
		mStatus.setName(id, me, name);
	}
	mStatus.setColor(id, me, color);
	mStatusDirty = true;

	if (me)
	{
		mConnectionState = STATE_CONNECTED;
		mMyID = id;

		if (mMyID == mMasterID)
		{
			activateMaster();
		}
		else if (mMasterID != GRAPPLE_SERVER)
		{
			mSandboxProxy->sendStage(mMasterID);
		}
	}
	else
	{
		const PlayerMap& map = mStatus.playerMap();

		PlayerMap::const_iterator iter = map.find(mMyID);

		if (iter != map.end())
		{
			mSandboxProxy->changeState(id, iter->second.time, iter->second.state);
		}
	}
}

void SandboxGame::userName(grapple_user id, const char* name)
{
	mStatus.setName(id, id == mMyID, name);
}

void SandboxGame::userDisconnected(grapple_user id)
{
	mStatus.erase(id);
}

void SandboxGame::serverDisconnected()
{
	mConnectionState = STATE_DISCONNECTED;
	mMaster = false;
	gotoLobby();
}

void SandboxGame::connectionRefused(grapple_connection_refused reason)
{
	mConnectionState = STATE_REFUSED;
	mReason = reason;
}

void SandboxGame::finishedGame()
{
	Main::pushState(gScreenEnd);
	changeMyState(PlayerInfo::STATE_FINISHED);
}

void SandboxGame::activateMaster()
{
	mMaster = true;
}

void SandboxGame::setMasterUser(const grapple_user& sender,
		const grapple_user& master)
{
	if (sender == GRAPPLE_SERVER)
	{
		mMasterID = master;

		if (mMyID == mMasterID)
		{
			activateMaster();
		}
		else if (mMyID != GRAPPLE_EVERYONE)
		{
			mSandboxProxy->sendStage(mMasterID);
		}
	}
}

void SandboxGame::sendStage(const grapple_user& sender)
{
	mSandboxProxy->loadLevel(sender, mLevelLoader->name());
}

void SandboxGame::loadLevel(const grapple_user& sender, const String& name)
{
	log(String("loadLevel: ") + String::toString(sender) + " - " + name);
	if (sender == mMasterID)
	{
		int id = mLevelLoader->getIdByName(name);
		if (id >= 0)
		{
			Game::stop(NULL);
			mLevelLoader->loadLevelByID(id);
			Game::start(NULL, false);
			mStatusDirty = true;

			if (mSandboxState != STATE_LOBBY)
			{
				Main::pushState(gScreenSandboxLobby);
				changeMyState(PlayerInfo::STATE_LOBBY);
			}
		}
	}
	else
	{
		log("rejected");
	}
}

void SandboxGame::resumeGame(const grapple_user& sender, const int& session)
{
	if (sender == mMasterID)
	{
		Main::popState();
		mStatus.reset(session);
		mLastUpdate = -1;
	}
}

void SandboxGame::endGame(const grapple_user& sender)
{
	if (sender == mMasterID)
	{
		gotoLobby();
	}
}

void SandboxGame::changeState(const grapple_user& sender, const float& time,
    const int& newState)
{
	mStatus.changeStatusState(sender, time, newState);
}

void SandboxGame::newTry(const grapple_user& sender, const int& session,
    const float& time)
{
	mStatus.attemptNewTry(sender, session, time);
}

void SandboxGame::setPosition(const grapple_user& sender, const int& session,
		const float& time, const Vector3& pos, const Vector3& vel)
{
	mStatus.updatePosition(sender, session, time, pos, vel);
}

void SandboxGame::explode(const grapple_user& sender)
{
	mStatus.causeExplosion(sender);
}
